--[[
    Change log:
        01/23/2021 1.0.0: initial release
]]

-- parent frame and namespace for this addon
local grounded = CreateFrame("Frame","GroundedFrame",UIParent,"PortraitFrameTemplate")

local FRAME_WIDTH = 272 -- width of the main frame
local FRAME_MIN_HEIGHT = 245 -- minimum height of the main frame
local LIST_MODE = 1 -- mode for displaying a list of bindings and adding new ones
local BIND_MODE = 2 -- mode for defining a single spell's binding
local CONFIRM_MODE = 3 -- mode for confirming something (binding override, unbinding)

GroundedSavedBinds = {} -- per-character savedvar of bindings

-- setting up window basics so it can join layout.txt for positioning
grounded:SetSize(272,250)
grounded:SetPoint("CENTER")
grounded:SetMovable(true)
grounded:Hide()

-- slash command to toggle window (alternative to button in spellbook)
SLASH_GROUNDED1 = "/grounded"
SlashCmdList["GROUNDED"] = function(msg)
    grounded:SetShown(not grounded:IsVisible())
end

--[[ events ]]

grounded:SetScript("OnEvent",function(self,event,...)
    if self[event] then
        self[event](self,...)
    end
end)
grounded:RegisterEvent("PLAYER_LOGIN")

-- on login do setup stuff
function grounded:PLAYER_LOGIN()
    -- savedvar housekeeping
    if type(GroundedSavedBinds)~="table" then
        GroundedSavedBinds = {}
    end
    self.savedBinds = GroundedSavedBinds
    -- finish setting up frame
    self:SetClampedToScreen(true)
    self:SetToplevel(true)
    self:SetScript("OnMouseDown",self.StartMoving)
    self:SetScript("OnMouseUp",self.StopMovingOrSizing)
    self:SetScript("OnShow",self.OnShow)
    self:SetScript("OnHide",self.OnHide)
    -- titlebar stuff
    self.portrait:SetTexture("Interface\\AddOns\\Grounded\\textures")
    self.portrait:SetTexCoord(0,0.5,0,0.5)
    self.TitleText:SetText("Grounded")
    self.TitleText:ClearAllPoints()
    self.TitleText:SetPoint("TOPLEFT",32,-6)
    self.TitleText:SetPoint("TOPRIGHT",0,-6)
    -- combat warning at top of window
    self.combatWarning = self:CreateFontString(nil,"ARTWORK","GameFontHighlightSmall")
    self.combatWarning:SetPoint("TOPLEFT",64,-24)
    self.combatWarning:SetPoint("BOTTOMRIGHT",self,"TOPRIGHT",-12,-64)
    self.combatWarning:SetText("You are in combat. Any changes will take effect when combat ends.")
    self.combatWarning:SetTextColor(1,0.82,0.5)
    -- start out listing binds
    self.mode = LIST_MODE
    -- setup insets
    self.listInset:Setup()
    self.bindInset:Setup()
    -- add Grounded button to the spellbook
    self:CreateToggleButton()
    -- finally, start up any existing saved bindings
    self:UpdateSecureButtons()
end

-- listInset wants to know when something was either picked up or dropped from the cursor
-- (this is only active when window on screen)
function grounded:CURSOR_UPDATE()
    self.listInset:UpdateCursor()
end

-- entering combat (this is only active when window on screen)
function grounded:PLAYER_REGEN_DISABLED()
    self.combatWarning:Show()
end

-- leaving combat (this is only active when window on screen)
function grounded:PLAYER_REGEN_ENABLED()
    self.combatWarning:Hide()
end

--[[ ui bits to setup for spellbook toggle button and parent GroundedFrame ]]

-- creates a little "tracking" button in topright corner of spellbook to toggle window
function grounded:CreateToggleButton()
    self.toggleButton = CreateFrame("Button","GroundedToggleButton",SpellBookFrame)
    self.toggleButton:SetSize(32,32)
    if WOW_PROJECT_ID == WOW_PROJECT_CLASSIC then -- on classic clients, move button in further
        self.toggleButton:SetPoint("TOPRIGHT",-44,-42)
    else
        self.toggleButton:SetPoint("TOPRIGHT",-24,-36)
    end
    self.toggleButton:SetNormalTexture("Interface\\AddOns\\Grounded\\textures")
    self.toggleButton:GetNormalTexture():SetTexCoord(0.5,1,0.5,1)
    self.toggleButton:SetPushedTexture("Interface\\AddOns\\Grounded\\textures")
    self.toggleButton:GetPushedTexture():SetTexCoord(0,0.5,0.5,1)
    self.toggleButton:SetHighlightTexture("Interface\\Minimap\\UI-Minimap-ZoomButton-Highlight")
    self.toggleButton:SetScript("OnClick",function(self)
        grounded:SetShown(not grounded:IsVisible())
    end)
    self.toggleButton:SetScript("OnEnter",function(self)
        GameTooltip:SetOwner(self,"ANCHOR_RIGHT")
        GameTooltip:AddLine("Grounded")
        GameTooltip:Show()
    end)
    self.toggleButton:SetScript("OnLeave",function(self)
        GameTooltip:Hide()
    end)
    hooksecurefunc("SpellBookFrame_Update",function(self)
        grounded.toggleButton:SetShown(SpellBookFrame.bookType=="spell")
    end)
end

-- when switching modes or need to update the frame, call this
function grounded:Update()
    self.combatWarning:SetShown(InCombatLockdown())
    self.listInset:SetShown(self.mode==LIST_MODE)
    self.bindInset:SetShown(self.mode==BIND_MODE)
    if self.mode==LIST_MODE then
        self.listInset:Update()
    elseif self.mode==BIND_MODE then
        self.bindInset:Update()
    end
    -- since there should never be more than half a dozen of these (maybe?), and we want to keep the "Place a ground-targeted" new
    -- button visible at all times, going to skip the scrollframe and just expand the window size
    self:SetHeight(max(FRAME_MIN_HEIGHT,self.listInset.numButtons*48+93))
end

-- when main frame is shown, always start in list mode and watch for combat status changing to warn about delayed saves
function grounded:OnShow()
    self:ShowBindList()
    self:RegisterEvent("PLAYER_REGEN_DISABLED")
    self:RegisterEvent("PLAYER_REGEN_ENABLED")
end

-- no need to watch for combat to update UI while it's hidden (secure stuff will handle its own combat awareness)
function grounded:OnHide()
    self:UnregisterEvent("PLAYER_REGEN_DISABLED")
    self:UnregisterEvent("PLAYER_REGEN_ENABLED")
end

-- changes frame to display the list of binds and the "Place a ground-targeted" new button
function grounded:ShowBindList()
    self.mode = LIST_MODE
    self:Update()
    -- if we've created, updated or deleted a bind it will come through here
    self:UpdateSecureButtons()
end

-- switch to bind mode for the given spell
function grounded:SetBind(spellName)
    if spellName then
        self.bindInset:SetBind(spellName)
        self.mode = BIND_MODE
        self:Update()
    end
end

--[[ secure stuff: where the magic happens ]]

local secureButtons = {} -- button pool of secure buttons

-- returns an available secure button, creating a new one if needed
function grounded:GetSecureButton()
    for _,button in ipairs(secureButtons) do
        if not button.isUsed then
            button.isUsed = true
            return button
        end
    end
    -- if reached here, no available existing button, create a new one
    local buttonName = "GroundedSecureButton"..(#secureButtons)
    local button = CreateFrame("Button",buttonName,nil,"SecureActionButtonTemplate")
    button:RegisterForClicks("AnyDown","AnyUp")
    button:SetAttribute("type","macro")
    -- on key down, cast normally so reticle appears; on key up, cast @cursor
    SecureHandlerWrapScript(button,"OnClick",button,[[
        local spellName = self:GetAttribute("spell") or ""
        if down then
            self:SetAttribute("macrotext","/cast "..spellName)
        else
            self:SetAttribute("macrotext","/stopspelltarget\n/cast [@cursor] "..spellName)
        end
    ]])
    -- add it to the pool
    button.isUsed = true
    tinsert(secureButtons,button)
    return button -- and return it
end

-- creates/updates buttons for each defined bind and sets it to the spell and key from savedvars
function grounded:UpdateSecureButtons()
    -- if attempting to update bindings during combat, setup a frame to wait until combat ends
    -- and try again. (note this isn't using the main PLAYER_REGEN_ENABLED event because that
    -- event is only active while the window on screen. The goal is for no events to be registered
    -- beyond PLAYER_LOGIN in the 99.99% of sessions that this window is never summoned.
    if InCombatLockdown() then
        if not grounded.inCombatFrame then
            grounded.inCombatFrame = grounded.inCombatFrame or CreateFrame("Frame")
            grounded.inCombatFrame:SetScript("OnEvent",function(self)
                self:UnregisterEvent("PLAYER_REGEN_ENABLED")
                grounded:UpdateSecureButtons()
            end)
        end
        -- register to listen for combat ending
        grounded.inCombatFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
        return -- and leave; we'll be back when combat ends
    end
    -- clear any existing bindings
    for _,button in ipairs(secureButtons) do
        ClearOverrideBindings(button)
        button.isUsed = false
    end
    -- go through and set all bindings, creating buttons as needed
    for index,bind in ipairs(grounded.savedBinds) do
        local button = grounded:GetSecureButton()
        button:SetAttribute("spell",bind[1])
        SetOverrideBindingClick(button,true,bind[2],button:GetName())
    end
end

--[[ listInset: the list of bindings ]]

grounded.listInset = CreateFrame("Frame",nil,grounded,"InsetFrameTemplate")

-- sets up the listInset UI
function grounded.listInset:Setup()
    self:SetPoint("TOPLEFT",4,-60)
    self:SetPoint("BOTTOMRIGHT",-6,28)
    self.buttons = {} -- ordered list of buttons being used in the list
    self:SetScript("OnShow",self.OnShow)
    self:SetScript("OnHide",self.OnHide)
    self:CreateOkButton()
end

-- creates the red Okay panel button in the bottomright
function grounded.listInset:CreateOkButton()
    local button = CreateFrame("Button",nil,self,"UIPanelButtonTemplate")
    button:SetWidth(self:GetWidth()/3)
    button:SetPoint("TOPRIGHT",self,"BOTTOMRIGHT",0,-1)
    button:SetText(OKAY)
    button:SetScript("OnClick",function(self) grounded:Hide() end)
end

-- updates the ui of the listInset and populates listInset.buttons in the process
function grounded.listInset:Update()
    grounded.listButtons:ReleaseButtons()
    wipe(self.buttons)
    self.numButtons = 0
    for _,bind in ipairs(grounded.savedBinds) do
        self.numButtons = self.numButtons + 1
        local button = grounded.listButtons:GetButton(self)
        button:SetPoint("TOPLEFT",4,-4-((self.numButtons-1)*48))
        button.bind = bind
        self:FillButton(button)
        tinsert(self.buttons,button)
    end
    -- last button is always the "Place a ground-targeted spell here" new button
    self.numButtons = self.numButtons + 1
    local button = grounded.listButtons:GetButton(self)
    button:SetPoint("TOPLEFT",4,-4-((self.numButtons-1)*48))
    button.bind = nil
    self:FillButton(button)
    tinsert(self.buttons,button)
end

-- updates contents of a list button; button.bind is {spellName,bindKey} when it's a binding, nil when it's a "Place a ground-targeted etc" new button
function grounded.listInset:FillButton(button)
    if button.bind then
        button.icon:SetTexture((GetSpellTexture(button.bind[1])) or "Interface\\Icons\\INV_Misc_QuestionMark")
        button.instructionsText:Hide()
        button.spellText:SetText(format("\124cffffd200Spell:\124r %s",button.bind[1]))
        button.bindText:SetText(format("\124cffffd200Bind:\124r %s",button.bind[2]))
        button.spellText:Show()
        button.bindText:Show()
    else
        button.icon:SetTexture("Interface\\PaperDollInfoFrame\\Character-Plus")
        button.instructionsText:Show()
        button.spellText:Hide()
        button.bindText:Hide()
    end
end

-- only watch for CURSOR_UPDATE when listInset is shown
function grounded.listInset:OnShow()
    grounded:RegisterEvent("CURSOR_UPDATE")
    self:UpdateCursor() -- in case a spell was on cursor when going to this inset
end

-- stop watching for CURSOR_UPDATE when listInset goes away
function grounded.listInset:OnHide()
    grounded:UnregisterEvent("CURSOR_UPDATE")
end

-- called on CURSOR_UPDATE, but contents of cursor aren't available in this execution path (grr), so come back in 1 frame
function grounded.listInset:UpdateCursor()
    C_Timer.After(0,grounded.listInset.UpdateFlash)
end

-- when a spell is picked up on the cursor, flash the last button in the list; stop flash when spell is dropped
function grounded.listInset:UpdateFlash()
    local cursorType = GetCursorInfo()
    local button = grounded.listInset.buttons[#grounded.listInset.buttons]
    if button then
        button.flash:SetShown(cursorType=="spell")
    end
end

--[[ listButtons: factory for list buttons used in listInset ]]

grounded.listButtons = {}

local buttonPool = {} -- simple button pool; use GetButton to get one and ReleaseButtons to release all

-- gets a listbutton from the pool, creating one if needed
function grounded.listButtons:GetButton(parent)
    for _,button in ipairs(buttonPool) do
        if not button.isUsed then
            button.isUsed = true
            button:Show()
            return button
        end
    end
    -- if reached here, need to make a new button
    local button = self:CreateButton(parent)
    button.isUsed = true
    tinsert(buttonPool,button)
    return button
end

-- puts all listbuttons back in the available pool
function grounded.listButtons:ReleaseButtons()
    for _,button in ipairs(buttonPool) do
        button.isUsed = false
        button:Hide()
    end
end

-- creates a list button (why did I not do this in XML)
function grounded.listButtons:CreateButton(parent)
    local button = CreateFrame("Button",nil,parent)
    button:SetSize(parent:GetWidth()-8,48)
    -- background and highlight textures
    self:CreateListButtonLayer(button,"back","BACKGROUND",0.126953125,0.173828125) -- background
    self:CreateListButtonLayer(button,"highlight","HIGHLIGHT",0.173828125,0.220703125) -- highlight
    -- animated border flash
    self:CreateListButtonFlash(button)
    -- icon on left side of button
    self:CreateListButtonIcon(button)
    -- spellText, bindText and instructionsText that take up most of the button
    self:CreateListButtonText(button)
    -- script handlers
    button:SetScript("OnMouseDown",self.ButtonOnMouseDown)
    button:SetScript("OnMouseUp",self.ButtonOnMouseUp)
    button:SetScript("OnShow",self.ButtonOnShow)
    button:SetScript("OnClick",self.ButtonOnClick)
    button:SetScript("OnReceiveDrag",self.ButtonOnClick)
    return button
end

-- to cut down on repition of code, creates a part of the background texture for list buttons
function grounded.listButtons:CreateListButtonPart(button,drawLayer,cx,cy,left,right,top,bottom,anchor)
    local part = button:CreateTexture(nil,drawLayer)
    part:SetSize(cx,cy)
    part:SetTexture("Interface\\PetBattles\\PetJournal")
    part:SetTexCoord(left,right,top,bottom)
    return part
end

-- creates the "back" (all black), "highlight" (blue border) or "border" (yellow border) layers for list buttons
function grounded.listButtons:CreateListButtonLayer(button,key,drawLayer,top,bottom)
    button[key] = {}
    button[key].left = self:CreateListButtonPart(button,drawLayer,48,48,0.49609375,0.58984375,top,bottom)
    button[key].left:SetPoint("LEFT")
    button[key].right = self:CreateListButtonPart(button,drawLayer,48,48,0.814453125,0.908203125,top,bottom)
    button[key].right:SetPoint("RIGHT")
    button[key].mid = self:CreateListButtonPart(button,drawLayer,0,48,0.58984375,0.814453125,top,bottom)
    button[key].mid:SetPoint("TOPLEFT",button[key].left,"TOPRIGHT")
    button[key].mid:SetPoint("BOTTOMRIGHT",button[key].right,"BOTTOMLEFT")
end

-- adds an animated pulsing yellow border to a list button (button.flash:Show() to start animation)
function grounded.listButtons:CreateListButtonFlash(button)
    button.flash = CreateFrame("Frame",nil,button)
    button.flash:SetAllPoints(button)
    button.flash:Hide()
    self:CreateListButtonLayer(button.flash,"border","BORDER",0.220703125,0.267578125)

    button.flash.animation = button.flash:CreateAnimationGroup()
    button.flash.alpha = button.flash.animation:CreateAnimation("Alpha")
    button.flash.alpha:SetDuration(0.5)
    button.flash.alpha:SetFromAlpha(0.5)
    button.flash.alpha:SetToAlpha(1)
    button.flash.animation:SetLooping("BOUNCE")

    button.flash:SetScript("OnShow",function(self) self.animation:Play() end)
    button.flash:SetScript("OnHide",function(self) self.animation:Stop() end)
end

-- adds an icon and a border around the icon to a list button
function grounded.listButtons:CreateListButtonIcon(button)
    button.icon = button:CreateTexture(nil,"BORDER")
    button.icon:SetSize(38,38)
    button.icon:SetPoint("LEFT",4,0)
    button.icon:SetTexCoord(0.075,0.925,0.075,0.925)
    button.icon:SetTexture("Interface\\PaperDollInfoFrame\\Character-Plus")

    button.iconBorder = button:CreateTexture(nil,"ARTWORK")
    button.iconBorder:SetSize(40,40)
    button.iconBorder:SetPoint("TOPLEFT",button.icon,"TOPLEFT",-1,1)
    button.iconBorder:SetTexture("Interface\\Buttons\\UI-QuickSlot2")
    button.iconBorder:SetTexCoord(0.1875,0.796875,0.1875,0.796875)
end

-- adds text to a list button
function grounded.listButtons:CreateListButtonText(button)
    button.spellText = button:CreateFontString(nil,"ARTWORK","GameFontHighlight")
    button.spellText:SetJustifyH("LEFT")
    button.spellText:SetPoint("TOPLEFT",button.icon,"TOPRIGHT",4,0)
    button.spellText:SetPoint("BOTTOMRIGHT",button,"RIGHT",-4,0)

    button.bindText = button:CreateFontString(nil,"ARTWORK","GameFontHighlight")
    button.bindText:SetJustifyH("LEFT")
    button.bindText:SetPoint("TOPLEFT",button.icon,"RIGHT",4,0)
    button.bindText:SetPoint("BOTTOMRIGHT",button,"BOTTOMRIGHT",-4,5)

    button.instructionsText = button:CreateFontString(nil,"ARTWORK","GameFontHighlight")
    button.instructionsText:SetJustifyH("LEFT")
    button.instructionsText:SetPoint("TOPLEFT",button.icon,"TOPRIGHT",4,0)
    button.instructionsText:SetPoint("BOTTOMRIGHT",button,"BOTTOMRIGHT",-4,5)
    button.instructionsText:SetText("Place a ground-targeted spell here to create a new key bind")
    button.instructionsText:SetTextColor(0.75,0.75,0.75)
end

-- when mouse goes down, save position of button and then nudge it -1,-2
function grounded.listButtons:ButtonOnMouseDown()
    self.anchorPoint,self.relativeTo,self.relativePoint,self.xoff,self.yoff = self:GetPoint(1)
    self:SetPoint(self.anchorPoint,self.relativeTo,self.relativePoint,self.xoff-1,self.yoff-2)
    self.isDown = true
end

-- when mouse goes up, restore to saved position
function grounded.listButtons:ButtonOnMouseUp()
    self:SetPoint(self.anchorPoint,self.relativeTo,self.relativePoint,self.xoff,self.yoff)
    self.isDown = false
end

-- in case button was "pushed" when it was hidden and won't get a MouseUp, restore its positon when shown again
function grounded.listButtons:ButtonOnShow()
    if self.isDown and self.xoff then
        self:SetPoint(self.anchorPoint,self.relativeTo,self.relativePoint,self.xoff,self.yoff)
    end
end

-- click of a "Place a ground-targeted etc" new button will go into bind mode if a spell is on cursor;
-- and click of an existing bind will go into bind mode for that bind
function grounded.listButtons:ButtonOnClick()
    local cursorType,_,_,spellID = GetCursorInfo()

    if not self.bind then
        if cursorType~="spell" then
            return -- need something on the cursor to bind
        end
        local spellName = GetSpellInfo(spellID)
        -- some covenant spells don't support @cursor unless you include their subtext, such as
        -- /cast [@cursor] Wild Spirits(Night Fae). on non-classic clients, add all subtexts just to
        -- be safe. (don't do this on classic because then it would always use the rank of the
        -- spell when the bind was made.)
        if WOW_PROJECT_ID ~= WOW_PROJECT_CLASSIC then
            local subtext = GetSpellSubtext(spellID)
            if subtext and subtext~="" then
                spellName = spellName.."("..subtext..")"
            end
        end
        ClearCursor()
        grounded:SetBind(spellName)
    else -- bind[1] is spellName and bind[2] is bindKey
        grounded:SetBind(self.bind[1])
    end
end

--[[ bindInset: binding a spell to a key ]]

grounded.bindInset = CreateFrame("Frame",nil,grounded,"InsetFrameTemplate")

-- does one-time setup of bindInset
function grounded.bindInset:Setup()
    self:SetPoint("TOPLEFT",4,-60)
    self:SetPoint("BOTTOMRIGHT",-6,28)
    self:CreateBindInsetUI()
    self:CreateCancelButton()
    self:CreateSaveButton()
    self:CreateUnbindButton()
    self.keyCapture = CreateFrame("Button",nil,self)
    self.keyCapture:SetScript("OnKeyDown",self.KeyCaptureOnKeyDown)
end

-- while bindInset is visible, the grounded.bindInset.keyCapture button will intercept all keys
function grounded.bindInset:KeyCaptureOnKeyDown(key)

    -- if ESC key hit, leave bind mode and return to list
    if key==GetBindingKey("TOGGLEGAMEMENU") then
        grounded:ShowBindList()
        return
    end

    if key=="LALT" or key=="RALT" or key=="ALT" then
        return -- do nothing if pressing ALT, wait for another key
    elseif key=="LSHIFT" or key=="RSHIFT" or key=="SHIFT" then
        return -- do nothing if pressing SHIFT, wait for another key
    elseif key=="LCTRL" or key=="RCTRL" or key=="CTRL" then
        return -- do nothing if pressing CTRL, wait for another key
    end

    -- if reached here, a non-modifier key was hit, attach modifiers
    local bind = IsAltKeyDown() and "ALT-" or ""
    bind = bind .. (IsControlKeyDown() and "CTRL-" or "")
    bind = bind .. (IsShiftKeyDown() and "SHIFT-" or "")
    bind = bind .. key

    -- and set bind
    self:GetParent().bindKey = bind
    self:GetParent():Update()
end

-- updates the bindInset UI
function grounded.bindInset:Update()
    self.spellText:SetText(self.spellName)
    self.icon:SetTexture(GetSpellTexture(self.spellName))
    self.bindText:SetText(self.bindKey or "(Undefined)")
    self.saveButton:SetEnabled(self.bindKey and true)
    self.unbindButton:SetEnabled(self.bindKey and true)
    self.warnText:Hide()
    local key = self.bindKey
    if key and GetBindingAction(key) then
        local action = GetBindingAction(key)
        -- if binding already exists as an action, warn that the override will override that action
        if action and action~="" then
            local actionName = _G["BINDING_NAME_"..action] or action
            self.warnText:SetText(format("\124TInterface\\DialogFrame\\DialogIcon-AlertNew-16:16\124t This will override the binding for %s.",actionName))
            self.warnText:Show()
        end
        -- don't allow [Enter], / or [Escape] keys (or whatever the open chat and game menu keys are) so user isn't "locked" out of doing stuff
        if key==GetBindingKey("OPENCHAT") or key==GetBindingKey("OPENCHATSLASH") or key==GetBindingKey("TOGGLEGAMEMENU") then
            self.warnText:SetText("\124TInterface\\DialogFrame\\DialogIcon-AlertNew-16:16\124t That key cannot be bound.")
            self.saveButton:Disable()
        end
    end
end

-- sets the spell we're about to bind; it will pick up the bindKey from savedvars if it already exists
function grounded.bindInset:SetBind(spellName)
    self.spellName = spellName
    self.bindKey = nil
    if spellName then
        for _,bind in ipairs(grounded.savedBinds) do
            if bind[1]==spellName then
                self.bindKey = bind[2]
            end
        end
    end
end

-- creates all the elements of the bindInset except for panel buttons
function grounded.bindInset:CreateBindInsetUI()
    -- name of the spell
    self.spellText = self:CreateFontString(nil,"ARTWORK","GameFontHighlight")
    self.spellText:SetPoint("TOP",22,-24)
    self.spellText:SetText("Rain of Fire")
    -- spell icon in topleft
    self.icon = self:CreateTexture(nil,"BORDER")
    self.icon:SetSize(32,32)
    self.icon:SetPoint("RIGHT",self.spellText,"LEFT",-8,0)
    self.icon:SetTexCoord(0.075,0.925,0.075,0.925)
    self.icon:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark")
    -- border around the icon
    self.iconBorder = self:CreateTexture(nil,"ARTWORK")
    self.iconBorder:SetSize(34,34)
    self.iconBorder:SetPoint("TOPLEFT",self.icon,"TOPLEFT",-1,1)
    self.iconBorder:SetTexture("Interface\\Buttons\\UI-QuickSlot2")
    self.iconBorder:SetTexCoord(0.1875,0.796875,0.1875,0.796875)
    -- "Press the key(s) to bind to this spell" instructions
    self.pressText = self:CreateFontString(nil,"ARTWORK","GameFontNormal")
    self.pressText:SetPoint("TOPLEFT",8,-58)
    self.pressText:SetPoint("TOPRIGHT",-8,-58)
    self.pressText:SetText("Press the key(s) to bind to this spell.")
    -- the binding keys text ("SHIFT-C" or "F12" etc)
    self.bindText = self:CreateFontString(nil,"ARTWORK","GameFontHighlightHuge")
    self.bindText:SetPoint("TOP",self.pressText,"BOTTOM",0,-10)
    self.bindText:SetText("SHIFT-C")
    -- warning text
    self.warnText = self:CreateFontString(nil,"ARTWORK","GameFontNormal")
    self.warnText:SetPoint("TOPLEFT",8,-108)
    self.warnText:SetPoint("TOPRIGHT",-8,-108)
    self.warnText:SetText("\124TInterface\\DialogFrame\\DialogIcon-AlertNew-16:16\124t SHIFT-C is currently bound to VERY_LONG_ACTION_NAME.")
end

-- creates Cancel button in the bottomright to return to bind list
function grounded.bindInset:CreateCancelButton()
    self.cancelButton = CreateFrame("Button",nil,self,"UIPanelButtonTemplate")
    self.cancelButton:SetWidth(self:GetWidth()/3)
    self.cancelButton:SetPoint("TOPRIGHT",self,"BOTTOMRIGHT",0,-1)
    self.cancelButton:SetText(CANCEL)
    self.cancelButton:SetScript("OnClick",function(self) grounded:ShowBindList() end)
end

-- creates Save button in bottom center to save displayed binding
function grounded.bindInset:CreateSaveButton()
    self.saveButton = CreateFrame("Button",nil,self,"UIPanelButtonTemplate")
    self.saveButton:SetWidth(self:GetWidth()/3)
    self.saveButton:SetPoint("TOP",self,"BOTTOM",0,-1)
    self.saveButton:SetText(SAVE)
    self.saveButton:SetScript("OnClick",function(self) grounded.bindInset:SaveBind() end)
end

-- creates Unbind button in bottomleft to delete displayed binding
function grounded.bindInset:CreateUnbindButton()
    self.unbindButton = CreateFrame("Button",nil,self,"UIPanelButtonTemplate")
    self.unbindButton:SetWidth(self:GetWidth()/3)
    self.unbindButton:SetPoint("TOPLEFT",self,"BOTTOMLEFT",0,-1)
    self.unbindButton:SetText("Unbind")
    self.unbindButton:SetScript("OnClick",function(self) grounded.bindInset:Unbind() end)
end

-- when Save button clicked, update an existing spell's bind or create a new bind in savedvars
function grounded.bindInset:SaveBind()
    if self.spellName and self.bindKey then
        for _,bind in ipairs(grounded.savedBinds) do
            if bind[1]==self.spellName then
                bind[2] = self.bindKey
                grounded:ShowBindList()
                return
            end
        end
        -- if reached here, no existing spell was bound, create a new one
        tinsert(grounded.savedBinds,{self.spellName,self.bindKey})
        grounded:ShowBindList()
    end
end

-- when Unbind button click, remove the displayed bind from savedvars and return to bind list
function grounded.bindInset:Unbind()
    for index,bind in ipairs(grounded.savedBinds) do
        if bind[1]==self.spellName then
            tremove(grounded.savedBinds,index)
            grounded:ShowBindList()
            return
        end
    end
    -- didn't find it, leave anyway
    grounded:ShowBindList()
end
